import { Transaction } from "@solana/web3.js";
import BN from "bn.js";
import { Context, PDA } from "..";

export class StakingClient {
  ctx: Context;
  public pda: PDA;

  constructor(ctx: Context, pda: PDA) {
    this.ctx = ctx;
    this.pda = pda;
  }

  public async stake(amount: BN) {
    const tx = new Transaction();

    const flexibleAccount = this.pda.flexibleStakingAccount(this.ctx.wallet.publicKey);
    const flexibleAccountInfo = await this.ctx.connection.getAccountInfo(flexibleAccount.key);
    if (!flexibleAccountInfo) {
      const initFlexibleAccountIx = await this.ctx.methods.initFlexibleAccount({
        accounts: {
          user: this.ctx.wallet.publicKey,
        },
        inputs: {
          bump: flexibleAccount.bump,
        },
      });
      if (!initFlexibleAccountIx.ix) {
        throw new Error("Failed to create initializing flexible account instruction");
      }
      tx.add(initFlexibleAccountIx.ix?.instructions[0]);
    }

    const stakeIx = await this.ctx.methods.stake({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        amount,
      },
    });
    if (!stakeIx.ix) {
      throw new Error("Failed to create staking instruction");
    }
    tx.add(stakeIx.ix?.instructions[0]);

    return tx;
  }

  public async withdraw(amount: BN, index: number) {
    const tx = new Transaction();

    const flexibleAccount = this.pda.flexibleStakingAccount(this.ctx.wallet.publicKey);
    const flexibleAccountInfo = await this.ctx.connection.getAccountInfo(flexibleAccount.key);
    if (!flexibleAccountInfo) {
      throw new Error("Ensure user has staked before withdrawing");
    }

    const withdrawPosition = this.pda.withdrawPosition(this.ctx.wallet.publicKey, new BN(index));
    const withdrawPositionInfo = await this.ctx.connection.getAccountInfo(withdrawPosition.key);
    if (withdrawPositionInfo) {
      throw new Error("Withdraw position index is already taken");
    }

    const withdrawIx = await this.ctx.methods.withdraw({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        withdrawPositionBump: withdrawPosition.bump,
        index,
        amount,
      },
    });
    if (!withdrawIx.ix) {
      throw new Error("Failed to create withdrawing instruction");
    }
    tx.add(withdrawIx.ix?.instructions[0]);

    return tx;
  }

  public async withdrawAfterDelay(index: number) {
    const tx = new Transaction();

    const withdrawPosition = this.pda.withdrawPosition(this.ctx.wallet.publicKey, new BN(index));
    const withdrawPositionInfo = await this.ctx.connection.getAccountInfo(withdrawPosition.key);
    if (!withdrawPositionInfo) {
      throw new Error("Ensure create withdraw position before getting delay amount");
    }

    const withdrawDelayIx = await this.ctx.methods.withdrawDelay({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        index,
      },
    });
    if (!withdrawDelayIx.ix) {
      throw new Error("Failed to create withdrawing after delay instruction");
    }
    tx.add(withdrawDelayIx.ix?.instructions[0]);

    return tx;
  }

  public async claim() {
    const tx = new Transaction();

    const flexibleAccount = this.pda.flexibleStakingAccount(this.ctx.wallet.publicKey);
    const flexibleAccountInfo = await this.ctx.connection.getAccountInfo(flexibleAccount.key);
    if (!flexibleAccountInfo) {
      throw new Error("Ensure user has staked before claiming");
    }

    const claimIx = await this.ctx.methods.claim({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
    });
    if (!claimIx.ix) {
      throw new Error("Failed to create claiming instruction");
    }
    tx.add(claimIx.ix?.instructions[0]);

    return tx;
  }

  public async lockup(amount: BN, index: number, vaultIndex: number) {
    const tx = new Transaction();

    const userLockupStakingAccount = this.pda.lockupStakingAccount(this.ctx.wallet.publicKey, new BN(index), new BN(vaultIndex));
    const userLockupStakingAccountInfo = await this.ctx.connection.getAccountInfo(userLockupStakingAccount.key);
    if (userLockupStakingAccountInfo) {
      throw new Error("Lockup staking account is already initialized");
    }

    const lockUpIx = await this.ctx.methods.lockup({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        userLockupStakingAccountBump: userLockupStakingAccount.bump,
        index,
        vaultIndex,
        amount,
      },
    });
    if (!lockUpIx.ix) {
      throw new Error("Failed to create locking up instruction");
    }
    tx.add(lockUpIx.ix?.instructions[0]);

    return tx;
  }

  public async unlock(index: number, vaultIndex: number) {
    const tx = new Transaction();

    const unlockIx = await this.ctx.methods.unlock({
      accounts: {
        user: this.ctx.wallet.publicKey,
      },
      inputs: {
        index,
        vaultIndex,
      },
    });
    if (!unlockIx.ix) {
      throw new Error("Failed to create unlocking instruction");
    }
    tx.add(unlockIx.ix?.instructions[0]);

    return tx;
  }
}